@@ -1,8 +1,11 @@
 use anchor_lang::solana_program::account_info::AccountInfo;
 use anchor_lang::solana_program::entrypoint::ProgramResult;
-use anchor_lang::{Accounts, CpiContext, ToAccountInfos};
+use anchor_lang::solana_program::program_error::ProgramError;
+use anchor_lang::solana_program::pubkey::Pubkey;
+use anchor_lang::{context::CpiContext, Accounts, ToAccountInfos};
 use serum_dex::instruction::SelfTradeBehavior;
 use serum_dex::matching::{OrderType, Side};
+use std::io::Write;
 use std::num::NonZeroU64;
 
 pub use serum_dex;
@@ -58,6 +61,30 @@ pub fn new_order_v3<'info>(
     Ok(())
 }
 
+pub fn cancel_order_v2<'info>(
+    ctx: CpiContext<'_, '_, '_, 'info, CancelOrderV2<'info>>,
+    side: Side,
+    order_id: u128,
+) -> ProgramResult {
+    let ix = serum_dex::instruction::cancel_order(
+        &ID,
+        ctx.accounts.market.key,
+        ctx.accounts.market_bids.key,
+        ctx.accounts.market_asks.key,
+        ctx.accounts.open_orders.key,
+        ctx.accounts.open_orders_authority.key,
+        ctx.accounts.event_queue.key,
+        side,
+        order_id,
+    )?;
+    solana_program::program::invoke_signed(
+        &ix,
+        &ToAccountInfos::to_account_infos(&ctx),
+        ctx.signer_seeds,
+    )?;
+    Ok(())
+}
+
 pub fn settle_funds<'info>(
     ctx: CpiContext<'_, '_, '_, 'info, SettleFunds<'info>>,
 ) -> ProgramResult {
@@ -91,6 +118,7 @@ pub fn init_open_orders<'info>(
         ctx.accounts.open_orders.key,
         ctx.accounts.authority.key,
         ctx.accounts.market.key,
+        ctx.remaining_accounts.first().map(|acc| acc.key),
     )?;
     solana_program::program::invoke_signed(
         &ix,
@@ -136,6 +164,41 @@ pub fn sweep_fees<'info>(ctx: CpiContext<'_, '_, '_, 'info, SweepFees<'info>>) -
     Ok(())
 }
 
+pub fn initialize_market<'info>(
+    ctx: CpiContext<'_, '_, '_, 'info, InitializeMarket<'info>>,
+    coin_lot_size: u64,
+    pc_lot_size: u64,
+    vault_signer_nonce: u64,
+    pc_dust_threshold: u64,
+) -> ProgramResult {
+    let authority = ctx.remaining_accounts.get(0);
+    let prune_authority = ctx.remaining_accounts.get(1);
+    let ix = serum_dex::instruction::initialize_market(
+        ctx.accounts.market.key,
+        &ID,
+        ctx.accounts.coin_mint.key,
+        ctx.accounts.pc_mint.key,
+        ctx.accounts.coin_vault.key,
+        ctx.accounts.pc_vault.key,
+        authority.map(|r| r.key),
+        prune_authority.map(|r| r.key),
+        ctx.accounts.bids.key,
+        ctx.accounts.asks.key,
+        ctx.accounts.req_q.key,
+        ctx.accounts.event_q.key,
+        coin_lot_size,
+        pc_lot_size,
+        vault_signer_nonce,
+        pc_dust_threshold,
+    )?;
+    solana_program::program::invoke_signed(
+        &ix,
+        &ToAccountInfos::to_account_infos(&ctx),
+        ctx.signer_seeds,
+    )?;
+    Ok(())
+}
+
 #[derive(Accounts)]
 pub struct NewOrderV3<'info> {
     pub market: AccountInfo<'info>,
@@ -158,6 +221,16 @@ pub struct NewOrderV3<'info> {
     pub rent: AccountInfo<'info>,
 }
 
+#[derive(Accounts)]
+pub struct CancelOrderV2<'info> {
+    pub market: AccountInfo<'info>,
+    pub market_bids: AccountInfo<'info>,
+    pub market_asks: AccountInfo<'info>,
+    pub open_orders: AccountInfo<'info>,
+    pub open_orders_authority: AccountInfo<'info>,
+    pub event_queue: AccountInfo<'info>,
+}
+
 #[derive(Accounts)]
 pub struct SettleFunds<'info> {
     pub market: AccountInfo<'info>,
@@ -171,6 +244,8 @@ pub struct SettleFunds<'info> {
     pub token_program: AccountInfo<'info>,
 }
 
+/// To use an (optional) market authority, add it as the first account of the
+/// CpiContext's `remaining_accounts` Vec.
 #[derive(Accounts)]
 pub struct InitOpenOrders<'info> {
     pub open_orders: AccountInfo<'info>,
@@ -196,3 +271,26 @@ pub struct SweepFees<'info> {
     pub vault_signer: AccountInfo<'info>,
     pub token_program: AccountInfo<'info>,
 }
+
+#[derive(Accounts)]
+pub struct InitializeMarket<'info> {
+    pub market: AccountInfo<'info>,
+    pub coin_mint: AccountInfo<'info>,
+    pub pc_mint: AccountInfo<'info>,
+    pub coin_vault: AccountInfo<'info>,
+    pub pc_vault: AccountInfo<'info>,
+    pub bids: AccountInfo<'info>,
+    pub asks: AccountInfo<'info>,
+    pub req_q: AccountInfo<'info>,
+    pub event_q: AccountInfo<'info>,
+    pub rent: AccountInfo<'info>,
+}
+
+#[derive(Clone)]
+pub struct Dex;
+
+impl anchor_lang::Id for Dex {
+    fn id() -> Pubkey {
+        ID
+    }
+}
